---
layout: docs
title: type_safe - Zero overhead abstractions for more type safety
project: type_safe
---

{{ standardese_doc $file $format }}

<hr class="footer"/><address class="footer"><small>
Documentation generated by <a target="_blank" href="https://github.com/foonathan/standardese">standardese</a>.
</small></address>
