// Copyright (C) 2016-2020 Jonathan Müller <jonathanmueller.dev@gmail.com>
// This file is subject to the license terms in the LICENSE file
// found in the top-level directory of this distribution.

#ifndef TYPE_SAFE_DETAIL_FORCE_INLINE_HPP_INCLUDED
#define TYPE_SAFE_DETAIL_FORCE_INLINE_HPP_INCLUDED

#include <debug_assert.hpp>

#define TYPE_SAFE_FORCE_INLINE DEBUG_ASSERT_FORCE_INLINE

#endif // TYPE_SAFE_DETAIL_FORCE_INLINE_HPP_INCLUDED
