/*
 * LibrePCB - Professional EDA for everyone!
 * Copyright (C) 2013 LibrePCB Developers, see AUTHORS.md for contributors.
 * https://librepcb.org/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef LIBREPCB_CORE_BOARDPICKPLACEGENERATOR_H
#define LIBREPCB_CORE_BOARDPICKPLACEGENERATOR_H

/*******************************************************************************
 *  Includes
 ******************************************************************************/
#include <librepcb/core/types/uuid.h>

#include <QtCore>

#include <memory>

/*******************************************************************************
 *  Namespace / Forward Declarations
 ******************************************************************************/
namespace librepcb {

class Board;
class PickPlaceData;

/*******************************************************************************
 *  Class BoardPickPlaceGenerator
 ******************************************************************************/

/**
 * @brief The BoardPickPlaceGenerator class
 */
class BoardPickPlaceGenerator final {
  Q_DECLARE_TR_FUNCTIONS(BoardPickPlaceGenerator)

public:
  // Constructors / Destructor
  BoardPickPlaceGenerator() = delete;
  BoardPickPlaceGenerator(const BoardPickPlaceGenerator& other) = delete;
  explicit BoardPickPlaceGenerator(const Board& board,
                                   const Uuid& assemblyVariant) noexcept;
  ~BoardPickPlaceGenerator() noexcept;

  // General Methods
  std::shared_ptr<PickPlaceData> generate() noexcept;

  // Operator Overloadings
  BoardPickPlaceGenerator& operator=(const BoardPickPlaceGenerator& rhs) =
      delete;

private:
  const Board& mBoard;
  const Uuid mAssemblyVariant;
};

/*******************************************************************************
 *  End of File
 ******************************************************************************/

}  // namespace librepcb

#endif
