#ifndef DXFLIB_RC
#define DXFLIB_RC

#include <windows.h>
#include "src/dl_dxf.h"
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US

#define VER_FILEVERSION             DL_VERSION_MAJOR,DL_VERSION_MINOR,DL_VERSION_REV,DL_VERSION_BUILD
#define VER_FILEVERSION_STR         DL_VERSION

#define VER_PRODUCTVERSION          DL_VERSION_MAJOR,DL_VERSION_MINOR,DL_VERSION_REV,DL_VERSION_BUILD
#define VER_PRODUCTVERSION_STR      DL_VERSION

#define VER_COMPANYNAME_STR         "RibbonSoft, GmbH"
#define VER_FILEDESCRIPTION_STR     "QCAD"
#define VER_INTERNALNAME_STR        "QCAD"
#define VER_LEGALCOPYRIGHT_STR      "Copyright (c) 2017 RibbonSoft, GmbH"
#define VER_LEGALTRADEMARKS1_STR    "All Rights Reserved"
#define VER_LEGALTRADEMARKS2_STR    VER_LEGALTRADEMARKS1_STR
#define VER_ORIGINALFILENAME_STR    "dxflib.dll"
#define VER_PRODUCTNAME_STR         "QCAD"

#define VER_COMPANYDOMAIN_STR       "https://qcad.org"

VS_VERSION_INFO VERSIONINFO
FILEVERSION VER_FILEVERSION
PRODUCTVERSION VER_PRODUCTVERSION
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904E4"
        BEGIN
            VALUE "CompanyName",        VER_COMPANYNAME_STR
            VALUE "FileDescription",    VER_FILEDESCRIPTION_STR
            VALUE "FileVersion",        VER_FILEVERSION_STR
            VALUE "InternalName",       VER_INTERNALNAME_STR
            VALUE "LegalCopyright",     VER_LEGALCOPYRIGHT_STR
            VALUE "LegalTrademarks1",   VER_LEGALTRADEMARKS1_STR
            VALUE "LegalTrademarks2",   VER_LEGALTRADEMARKS2_STR
            VALUE "OriginalFilename",   VER_ORIGINALFILENAME_STR
            VALUE "ProductName",        VER_PRODUCTNAME_STR
            VALUE "ProductVersion",     VER_PRODUCTVERSION_STR
            //VALUE "Rev",     I_SVNREV
            //VALUE "Date",     I_SVN_CLOCKDATE
        END
    END

    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1252
    END
END
#endif // DXFLIB_RC
