#include <QtCore>
#include "dimension.h"
#include "../common/domelement.h"

namespace parseagle {

Dimension::Dimension(const DomElement& root)
{
    Q_UNUSED(root);
}

Dimension::~Dimension() noexcept
{
}

} // namespace parseagle
