/*
 * LibrePCB - Professional EDA for everyone!
 * Copyright (C) 2013 LibrePCB Developers, see AUTHORS.md for contributors.
 * https://librepcb.org/
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef LIBREPCB_EDITOR_CMDATTRIBUTEEDIT_H
#define LIBREPCB_EDITOR_CMDATTRIBUTEEDIT_H

/*******************************************************************************
 *  Includes
 ******************************************************************************/
#include "../undocommand.h"
#include "cmdlistelementinsert.h"
#include "cmdlistelementremove.h"
#include "cmdlistelementsswap.h"

#include <librepcb/core/attribute/attribute.h>

#include <QtCore>

/*******************************************************************************
 *  Namespace / Forward Declarations
 ******************************************************************************/
namespace librepcb {
namespace editor {

/*******************************************************************************
 *  Class CmdAttributeEdit
 ******************************************************************************/

/**
 * @brief The CmdAttributeEdit class
 */
class CmdAttributeEdit final : public UndoCommand {
public:
  // Constructors / Destructor
  CmdAttributeEdit() = delete;
  CmdAttributeEdit(const CmdAttributeEdit& other) = delete;
  explicit CmdAttributeEdit(Attribute& attribute) noexcept;
  ~CmdAttributeEdit() noexcept;

  // Setters
  void setKey(const AttributeKey& key) noexcept;
  void setType(const AttributeType& type) noexcept;
  void setValue(const QString& value) noexcept;
  void setUnit(const AttributeUnit* unit) noexcept;

  // Operator Overloadings
  CmdAttributeEdit& operator=(const CmdAttributeEdit& rhs) = delete;

private:
  // Private Methods

  /// @copydoc ::librepcb::editor::UndoCommand::performExecute()
  bool performExecute() override;

  /// @copydoc ::librepcb::editor::UndoCommand::performUndo()
  void performUndo() override;

  /// @copydoc ::librepcb::editor::UndoCommand::performRedo()
  void performRedo() override;

  // Private Member Variables

  // Attributes from the constructor
  Attribute& mAttribute;

  // General Attributes
  AttributeKey mOldKey;
  AttributeKey mNewKey;
  const AttributeType* mOldType;
  const AttributeType* mNewType;
  QString mOldValue;
  QString mNewValue;
  const AttributeUnit* mOldUnit;
  const AttributeUnit* mNewUnit;
};

/*******************************************************************************
 *  Undo Commands
 ******************************************************************************/

using CmdAttributeInsert =
    CmdListElementInsert<Attribute, AttributeListNameProvider,
                         Attribute::Event>;
using CmdAttributeRemove =
    CmdListElementRemove<Attribute, AttributeListNameProvider,
                         Attribute::Event>;
using CmdAttributesSwap =
    CmdListElementsSwap<Attribute, AttributeListNameProvider, Attribute::Event>;

/*******************************************************************************
 *  End of File
 ******************************************************************************/

}  // namespace editor
}  // namespace librepcb

#endif
